/**************************************************************/
/***                                                        ***/
/***  Projeto de Geometria COmputacional - Mac 331          ***/
/***                                                        ***/
/***  Implementao do Algoritmo de Graham - Fecho Convexo  ***/
/***                                                        ***/
/***  Jacques Salo Zweiman            N.USP : 2237982       ***/
/***  Henrico Scaranello              N.USP : 2370348       ***/
/***                                                        ***/
/***  Professor : Jos Coelho Pina Jr.                      ***/
/***                                                        ***/
/**************************************************************/

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <stdio.h>
#include <math.h>

#define MIN_TEMPO 1000 // 1 milisegundo

typedef struct Ponto {
  float X;
  float Y;
} Ponto;


typedef struct Cor {
  float R;
  float G;
  float B;
} Cor;

typedef struct Item {
  Ponto p;
  struct Item *ant;
} Item;


char *nomearq;
Ponto *pontos;
Ponto *newpontos;
Ponto *fecho;
int nfecho;
int npontos;
int newnpontos;
int continua;
Cor branco;
Cor verde;
Cor azul;
Cor vermelho;
int minx, miny, maxx, maxy;
Item *topo;
int fator;
static int indice;

int CalculaTempo() {
  if (fator <= 0)
    return 0;	    
  return MIN_TEMPO * 1000 / fator;
}

void marcaPonto(Ponto a, Cor c) {
  glColor3f(c.R, c.G, c.B);
  glBegin(GL_POINTS);
  glVertex2f(a.X, a.Y);
  glEnd();
}

void marcaPontos(Cor c) {
  int i; 
  for(i = 0; i < npontos; i++)
    marcaPonto(pontos[i],c);
}

void marcaPontosFecho(Cor c) {
  Item *i;
  glColor3f(c.R, c.G, c.B);
  glBegin(GL_LINE_STRIP);
  for (i = topo; i != NULL; i = i->ant)
    glVertex2f(i->p.X, i->p.Y);
  glEnd();
}

void redesenha() {
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  marcaPontos(branco);
  marcaPontosFecho(verde);
  glutSwapBuffers();
  usleep(CalculaTempo());
}

void Tira() {
  topo = topo->ant;
  nfecho--;
}

Ponto Pega() {
  return topo->p;
}

Ponto Pega2() {
  return topo->ant->p;
}

void Insere(Ponto p) {
  Item *i;
  i = (Item *) malloc (sizeof(Item));
  i->p = p;
  i->ant = topo;
  topo = i;
  nfecho++;
}

float Area2(Ponto a, Ponto b, Ponto c) {
  return a.X * b.Y -  a.Y * b.X + a.Y * c.X - a.X * c.Y + b.X * c.Y - c.X * b.Y;
}

void tracaReta(Ponto a, Ponto b, Cor c) {
  glColor3f(c.R, c.G, c.B);
  glBegin(GL_LINES);
  glVertex2f(a.X, a.Y);
  glVertex2f(b.X, b.Y);
  glEnd();
}

int Left(Ponto a, Ponto b, Ponto c) {
  tracaReta(a,b,vermelho);
  tracaReta(b,c,azul);
  tracaReta(c,a,azul);      
  glutSwapBuffers();
  usleep(CalculaTempo());
  return (Area2(a,b,c) > 0.0);
}

Ponto menorY() {
  int i, min;
  Ponto aux;
  min = 0;
  for (i = 1; i < npontos; i++) {
    if (pontos[i].Y < pontos[min].Y)
      min = i;
    else 
      if ((pontos[i].Y == pontos[min].Y) && (pontos[min].X < pontos[i].X))
	min = i;
  }

  aux = pontos[min];
  for (i = min; i < (npontos-1); i++) {
    pontos[i] = pontos[i+1];
  }
  return(aux);
}

int equal(Ponto a, Ponto b) {
  return (a.X == b.X && a.Y == b.Y);
}

static Ponto p0;
float dist(Ponto a){
  return sqrt(pow((a.X - p0.X), 2) + pow((a.Y - p0.Y), 2));
}

float projx(Ponto a){
  return a.X - p0.X;
}

float coseno(Ponto a) {
  if (dist(a) == 0) 
    return 0;
  else
    return projx(a) / dist(a);
}

int compara(void *p1, void *p2) {
  float c1, c2;
  float d1, d2;
  c1 = coseno(*(Ponto *) p1);
  c2 = coseno(*(Ponto *) p2);

  if (c1 > c2) return 1;
  else if (c1 < c2) return -1;
  else {
    d1 = dist(*(Ponto *) p1);
    d2 = dist(*(Ponto *) p2);
    if (d1 > d2) return 1;
    else if (d1 < d2) return -1;
    else return 0;
  }
}

void grahamSort() {
  qsort(pontos, npontos-1, sizeof(Ponto), compara);
}

void inverteVetor() {
  int i;
  Ponto aux;
  for (i = 0; i < (npontos)/2; i++) {
    aux = pontos[npontos-i-2];
    pontos[npontos-i-1] = pontos[i];
    pontos[i] = aux;
  }
}

void eliminaRepeticoes() {
  int i, j;

  newpontos = (Ponto *) malloc((npontos + 1) * sizeof(Ponto));  

  newnpontos = npontos;
  i = 1;
  j = 1;
  newpontos[0] = pontos[0];
  while (i < npontos) {
    newpontos[j++] = pontos[i];
    while ((i<npontos-1) && (coseno(pontos[i]) == coseno(pontos[i+1]))) {
      i++;
      newnpontos--;
    }
    i++;
  }
}

void inigraham() {
  if(lePontos(nomearq))
    return;
  
  topo = NULL;
  redesenha();
  p0 = menorY();
  grahamSort();
  inverteVetor();
  eliminaRepeticoes();
}
  
void graham() {
  int i;

  nfecho = 0;
  Insere(p0);
  Insere(newpontos[0]);
  Insere(newpontos[1]);
  redesenha();
  
  indice = 2;
  if (fator == -1) {
    grahampasso(indice);
    return;
  }

  for (i = 2; i < newnpontos; ) {
    i += grahampasso(i);
  }
  Insere(p0);
  redesenha();
}

int grahampasso(int i) {
  int ret = 0;
  if (Left(Pega2(), Pega(), newpontos[i])) {
    Insere(newpontos[i]);
    if (fator == -1)
      indice ++;
    ret = 1;
  }
  else {
    Tira();
  }
  redesenha();
  return ret;
}

int lePontos(char *nome_arq) {
 FILE *entrada;
 int i;
 char c;

 entrada = fopen(nome_arq, "r");
 
 if(entrada == NULL) {
   printf("Erro na abertura do arquivo");
   return 1;
 }

 fscanf(entrada, "%d\n", &npontos);

 pontos = (Ponto *) malloc(npontos * sizeof(Ponto));
 fscanf(entrada, "%f %f", &pontos[0].X, &pontos[0].Y);
 minx = pontos[0].X;
 maxx = pontos[0].X;
 miny = pontos[0].Y;
 maxy = pontos[0].Y;
 for(i = 1; i < npontos; i++) {
   fscanf(entrada, "%f %f", &pontos[i].X, &pontos[i].Y);
   if (pontos[i].X > maxx)
     maxx = pontos[i].X;
   else
     if (pontos[i].X < minx)
       minx = pontos[i].X;
   if (pontos[i].Y > maxy)
     maxy = pontos[i].Y;
   else
     if (pontos[i].Y < miny)
       miny = pontos[i].Y;
 }

 fclose(entrada);
 return 0;
}


void init(void);
void display(void);

void key(unsigned char key, int x, int y) {
  switch (key) {
  case 'N':
  case 'n':
    if (indice < newnpontos)
      grahampasso(indice);
    else {
      Insere(p0);
      redesenha();
      fator = 16;
    }
    break;
  }
}
void menu(int v) {
  switch (v) {
  case 3:
    inigraham();
    break;
  case 5:
    graham();
    break;
  case 1: 
    fator = -1;
    break;
  case 0:
  case 2:
  case 4:  
  case 8:
  case 16:
  case 32:
  case 64:
  case 1000:
  case 1000000:
    fator = v;
    break;
  case 999:
    exit(1);
  }
}

int main (int argc, char **argv)
{
  int s;
  azul.R = 0.0;
  azul.G = 0.0;
  azul.B = 1.0;
  
  branco.R = 1.0;
  branco.G = 1.0;
  branco.B = 1.0;

  verde.R = 0.0;
  verde.G = 1.0;
  verde.B = 0.0;

  vermelho.R = 1.0;
  vermelho.G = 0.0;
  vermelho.B = 0.0;

  fator = 16;
  
  if(argc < 2) {
    printf("Utilizao: Graham <arq_entrada>\n");
    return (-1);
  }
  else {
    nomearq = argv[1];
    if(lePontos(nomearq))
      return;
  }

  // Inicializaes do openGL
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  
  glutInitWindowSize(300, 300);	
  glutInitWindowPosition(100, 100);
  glutCreateWindow("Algoritmo de Graham !!!");

  glClearColor(0.0, 0.0, 0.0, 0.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(minx-1, maxx+1, miny-1, maxy+1);

  s = glutCreateMenu(menu);
  glutAddMenuEntry("Rele pontos", 3);
  glutAddMenuEntry("Inicia Execuo", 5);
  glutAddMenuEntry("Passo a passo", 1);
  glutAddMenuEntry("1/2 segundo", 2);
  glutAddMenuEntry("1/4 segundo", 4);
  glutAddMenuEntry("1/8 segundo", 8);
  glutAddMenuEntry("1/16 segundo", 16);
  glutAddMenuEntry("1/32 segundo", 32);
  glutAddMenuEntry("1/64 segundo", 64);
  glutAddMenuEntry("Rpido", 1000);
  glutAddMenuEntry("Super-Rpido", 1000000);
  glutAddMenuEntry("Voando !!!", 0);
  glutAddMenuEntry("Sair", 999);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  glutKeyboardFunc(key);

  glutDisplayFunc(display);
  glutMainLoop();
  return 0;
}

void display(void)
{
  inigraham();
  glutSwapBuffers();
}
